import os
import sys
import warnings

import six

from dagster import check
from dagster.api.list_repositories import sync_list_repositories
from dagster.core.code_pointer import CodePointer, rebase_file
from dagster.core.host_representation import RepositoryLocationHandle, UserProcessApi
from dagster.grpc.types import LoadableTargetOrigin
from dagster.utils import load_yaml_from_path

from .config_schema import ensure_workspace_config
from .workspace import Workspace


def load_workspace_from_yaml_paths(yaml_paths, python_user_process_api):
    check.list_param(yaml_paths, 'yaml_paths', str)
    check.inst_param(python_user_process_api, 'python_user_process_api', UserProcessApi)

    workspace_configs = [load_yaml_from_path(yaml_path) for yaml_path in yaml_paths]
    workspaces = [
        load_workspace_from_config(workspace_config, yaml_path, python_user_process_api)
        for workspace_config, yaml_path in zip(workspace_configs, yaml_paths)
    ]

    repository_location_handles_dict = {}
    for workspace in workspaces:
        for repository_location_name in workspace.repository_location_names:
            repository_location_handles_dict[
                repository_location_name
            ] = workspace.get_repository_location_handle(repository_location_name)

    repository_location_handles = list(repository_location_handles_dict.values())
    merged_workspace = Workspace(repository_location_handles)
    return merged_workspace


def load_workspace_from_config(workspace_config, yaml_path, python_user_process_api):
    ensure_workspace_config(workspace_config, yaml_path)
    check.inst_param(python_user_process_api, 'python_user_process_api', UserProcessApi)

    if 'repository' in workspace_config:
        warnings.warn(
            # link to docs once they exist
            'You are using the legacy repository yaml format. Please update your file '
            'to abide by the new workspace file format.'
        )
        return Workspace(
            [
                RepositoryLocationHandle.create_in_process_location(
                    pointer=CodePointer.from_legacy_repository_yaml(yaml_path)
                )
            ]
        )

    location_handles = []
    for location_config in workspace_config['load_from']:
        location_handles.append(
            _location_handle_from_location_config(
                location_config, yaml_path, python_user_process_api
            )
        )

    return Workspace(location_handles)


def _location_handle_from_module_config(
    python_module_config, user_process_api, executable_path=sys.executable
):
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)
    module_name, attribute, location_name = _get_module_config_data(python_module_config)
    return location_handle_from_module_name(
        module_name, attribute, user_process_api, location_name, executable_path
    )


def _get_module_config_data(python_module_config):
    return (
        (python_module_config, None, None)
        if isinstance(python_module_config, six.string_types)
        else (
            python_module_config['module_name'],
            python_module_config.get('attribute'),
            python_module_config.get('location_name'),
        )
    )


def location_handle_from_module_name(
    module_name, attribute, user_process_api, location_name=None, executable_path=sys.executable
):
    check.str_param(module_name, 'module_name')
    check.opt_str_param(attribute, 'attribute')
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)
    check.opt_str_param(location_name, 'location_name')

    if user_process_api == UserProcessApi.GRPC:
        return RepositoryLocationHandle.create_process_bound_grpc_server_location(
            LoadableTargetOrigin(
                executable_path=executable_path,
                python_file=None,
                module_name=module_name,
                working_directory=None,
                attribute=attribute,
            ),
            location_name,
        )
    else:
        response = sync_list_repositories(
            executable_path=executable_path,
            python_file=None,
            module_name=module_name,
            working_directory=None,
            attribute=attribute,
        )
        return RepositoryLocationHandle.create_python_env_location(
            executable_path=executable_path,
            location_name=location_name,
            repository_code_pointer_dict={
                lrs.repository_name: CodePointer.from_module(module_name, lrs.attribute)
                for lrs in response.repository_symbols
            },
        )


def _location_handle_from_package_config(
    python_package_config, user_process_api, executable_path=sys.executable
):
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)
    module_name, attribute, location_name = _get_package_config_data(python_package_config)
    return location_handle_from_package_name(
        module_name, attribute, user_process_api, location_name, executable_path
    )


def _get_package_config_data(python_package_config):
    return (
        (python_package_config, None, None)
        if isinstance(python_package_config, six.string_types)
        else (
            python_package_config['package_name'],
            python_package_config.get('attribute'),
            python_package_config.get('location_name'),
        )
    )


def location_handle_from_package_name(
    package_name, attribute, user_process_api, location_name=None, executable_path=sys.executable
):
    check.str_param(package_name, 'package_name')
    check.opt_str_param(attribute, 'attribute')
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)
    check.opt_str_param(location_name, 'location_name')

    if user_process_api == UserProcessApi.GRPC:
        return RepositoryLocationHandle.create_process_bound_grpc_server_location(
            loadable_target_origin=LoadableTargetOrigin(
                executable_path=executable_path,
                python_file=None,
                module_name=package_name,
                working_directory=None,
                attribute=attribute,
            ),
            location_name=location_name,
        )
    else:
        response = sync_list_repositories(
            executable_path=executable_path,
            python_file=None,
            module_name=package_name,
            working_directory=None,
            attribute=attribute,
        )
        return RepositoryLocationHandle.create_python_env_location(
            executable_path=executable_path,
            location_name=location_name,
            repository_code_pointer_dict={
                lrs.repository_name: CodePointer.from_python_package(package_name, lrs.attribute)
                for lrs in response.repository_symbols
            },
        )


def _location_handle_from_python_file_config(
    python_file_config, yaml_path, user_process_api, executable_path=sys.executable
):
    check.str_param(yaml_path, 'yaml_path')
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)

    absolute_path, attribute, location_name, working_directory = _get_python_file_config_data(
        python_file_config, yaml_path
    )

    return location_handle_from_python_file(
        absolute_path,
        attribute,
        working_directory,
        user_process_api,
        location_name,
        executable_path,
    )


def _get_python_file_config_data(python_file_config, yaml_path):
    return (
        (rebase_file(python_file_config, yaml_path), None, None, None)
        if isinstance(python_file_config, six.string_types)
        else (
            rebase_file(python_file_config['relative_path'], yaml_path),
            python_file_config.get('attribute'),
            python_file_config.get('location_name'),
            rebase_file(python_file_config.get('working_directory'), yaml_path)
            if python_file_config.get('working_directory')
            else rebase_file(os.path.dirname(yaml_path), yaml_path),
        )
    )


def location_handle_from_python_file(
    python_file,
    attribute,
    working_directory,
    user_process_api,
    location_name=None,
    executable_path=sys.executable,
):
    check.str_param(python_file, 'python_file')
    check.opt_str_param(attribute, 'attribute')
    check.opt_str_param(working_directory, 'working_directory')
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)
    check.opt_str_param(location_name, 'location_name')

    if user_process_api == UserProcessApi.GRPC:
        return RepositoryLocationHandle.create_process_bound_grpc_server_location(
            loadable_target_origin=LoadableTargetOrigin(
                executable_path=executable_path,
                python_file=python_file,
                module_name=None,
                working_directory=working_directory,
                attribute=attribute,
            ),
            location_name=location_name,
        )
    else:
        response = sync_list_repositories(
            executable_path=executable_path,
            python_file=python_file,
            module_name=None,
            working_directory=working_directory,
            attribute=attribute,
        )
        return RepositoryLocationHandle.create_python_env_location(
            executable_path=executable_path,
            location_name=location_name,
            repository_code_pointer_dict={
                lrs.repository_name: CodePointer.from_python_file(
                    python_file, lrs.attribute, working_directory
                )
                for lrs in response.repository_symbols
            },
        )


def _location_handle_from_grpc_server_config(grpc_server_config, yaml_path):
    check.dict_param(grpc_server_config, 'grpc_server_config')
    check.str_param(yaml_path, 'yaml_path')

    port, socket, host, location_name = (
        grpc_server_config.get('port'),
        grpc_server_config.get('socket'),
        grpc_server_config.get('host'),
        grpc_server_config.get('location_name'),
    )

    check.invariant(
        (socket or port) and not (socket and port), "must supply either a socket or a port"
    )

    if not host:
        host = 'localhost'

    return RepositoryLocationHandle.create_grpc_server_location(
        port=port, socket=socket, host=host, location_name=location_name,
    )


def _location_handle_from_python_environment_config(
    python_environment_config, yaml_path, user_process_api
):
    check.dict_param(python_environment_config, 'python_environment_config')
    check.str_param(yaml_path, 'yaml_path')
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)

    executable_path, target_config = (
        # do shell expansion on path
        os.path.expanduser(python_environment_config['executable_path']),
        python_environment_config['target'],
    )

    check.invariant(is_target_config(target_config))

    python_file_config, python_module_config, python_package_config = (
        target_config.get('python_file'),
        target_config.get('python_module'),
        target_config.get('python_package'),
    )

    if python_file_config:
        return _location_handle_from_python_file_config(
            python_file_config, yaml_path, user_process_api, executable_path
        )
    elif python_module_config:
        return _location_handle_from_module_config(
            python_module_config, user_process_api, executable_path
        )
    else:
        check.invariant(python_package_config)
        return _location_handle_from_package_config(
            python_package_config, user_process_api, executable_path
        )


def _location_handle_from_location_config(location_config, yaml_path, python_user_process_api):
    check.dict_param(location_config, 'location_config')
    check.str_param(yaml_path, 'yaml_path')
    check.inst_param(python_user_process_api, 'python_user_process_api', UserProcessApi)

    if is_target_config(location_config):
        return _location_handle_from_target_config(
            location_config, yaml_path, python_user_process_api
        )

    elif 'grpc_server' in location_config:
        return _location_handle_from_grpc_server_config(location_config['grpc_server'], yaml_path)

    elif 'python_environment' in location_config:
        return _location_handle_from_python_environment_config(
            location_config['python_environment'], yaml_path, python_user_process_api
        )

    else:
        check.not_implemented('Unsupported location config: {}'.format(location_config))


def is_target_config(potential_target_config):
    return isinstance(potential_target_config, dict) and (
        potential_target_config.get('python_file')
        or potential_target_config.get('python_module')
        or potential_target_config.get('python_package')
    )


def _location_handle_from_target_config(target_config, yaml_path, user_process_api):
    check.dict_param(target_config, 'target_config')
    check.param_invariant(is_target_config(target_config), 'target_config')
    check.str_param(yaml_path, 'yaml_path')
    check.inst_param(user_process_api, 'user_process_api', UserProcessApi)

    if 'python_file' in target_config:
        return _location_handle_from_python_file_config(
            target_config['python_file'], yaml_path, user_process_api
        )

    elif 'python_module' in target_config:
        return _location_handle_from_module_config(target_config['python_module'], user_process_api)

    elif 'python_package' in target_config:
        return _location_handle_from_package_config(
            target_config['python_package'], user_process_api
        )

    else:
        check.failed('invalid target_config')



from __future__ import print_function

from contextlib import contextmanager
from textwrap import TextWrapper

from six import StringIO

from dagster import check

LINE_LENGTH = 100


class IndentingPrinter(object):
    def __init__(self, indent_level=2, printer=print, current_indent=0, line_length=LINE_LENGTH):
        self.current_indent = current_indent
        self.indent_level = check.int_param(indent_level, 'indent_level')
        self.printer = check.callable_param(printer, 'printer')
        self.line_length = line_length

        self._line_so_far = ''

    def append(self, text):
        check.str_param(text, 'text')
        self._line_so_far += text

    def line(self, text):
        check.str_param(text, 'text')
        self.printer(self.current_indent_str + self._line_so_far + text)
        self._line_so_far = ''

    def block(self, text, prefix='', initial_indent=''):
        '''Automagically wrap a block of text.'''
        wrapper = TextWrapper(
            width=self.line_length - len(self.current_indent_str),
            initial_indent=initial_indent,
            subsequent_indent=prefix,
            break_long_words=False,
            break_on_hyphens=False,
        )
        for line in wrapper.wrap(text):
            self.line(line)

    def comment(self, text):
        self.block(text, prefix='# ', initial_indent='# ')

    @property
    def current_indent_str(self):
        return ' ' * self.current_indent

    def blank_line(self):
        check.invariant(
            not self._line_so_far, 'Cannot throw away appended strings by calling blank_line'
        )
        self.printer('')

    def increase_indent(self):
        self.current_indent += self.indent_level

    def decrease_indent(self):
        if self.indent_level and self.current_indent <= 0:
            raise Exception('indent cannot be negative')
        self.current_indent -= self.indent_level

    @contextmanager
    def with_indent(self, text=None):
        if text is not None:
            self.line(text)
        self.increase_indent()
        yield
        self.decrease_indent()


class IndentingStringIoPrinter(IndentingPrinter):
    '''Subclass of IndentingPrinter wrapping a StringIO.'''

    def __init__(self, **kwargs):
        self.buffer = StringIO()
        self.printer = lambda x: self.buffer.write(x + '\n')
        super(IndentingStringIoPrinter, self).__init__(printer=self.printer, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, _exception_type, _exception_value, _traceback):
        self.buffer.close()

    def read(self):
        '''Get the value of the backing StringIO.'''
        return self.buffer.getvalue()



'''
This subpackage contains all classes that host processes (e.g. dagit)
use to manipulate and represent definitions that are resident
in user processes and containers.  e.g. ExternalPipeline

It also contains classes that represent historical representations
that have been persisted. e.g. HistoricalPipeline
'''
from .external import (
    ExternalExecutionPlan,
    ExternalPartitionSet,
    ExternalPipeline,
    ExternalPipelineExecutionResult,
    ExternalRepository,
    ExternalSchedule,
)
from .external_data import (
    ExternalPartitionConfigData,
    ExternalPartitionExecutionErrorData,
    ExternalPartitionNamesData,
    ExternalPartitionSetData,
    ExternalPartitionTagsData,
    ExternalPipelineData,
    ExternalPipelineSubsetResult,
    ExternalPresetData,
    ExternalRepositoryData,
    ExternalScheduleData,
    ExternalScheduleExecutionData,
    ExternalScheduleExecutionErrorData,
    external_pipeline_data_from_def,
    external_repository_data_from_def,
)
from .handle import (
    GrpcServerRepositoryLocationHandle,
    InProcessRepositoryLocationHandle,
    ManagedGrpcPythonEnvRepositoryLocationHandle,
    PipelineHandle,
    PythonEnvRepositoryLocationHandle,
    RepositoryHandle,
    RepositoryLocationHandle,
    UserProcessApi,
)
from .historical import HistoricalPipeline
from .pipeline_index import PipelineIndex
from .repository_location import (
    GrpcServerRepositoryLocation,
    InProcessRepositoryLocation,
    PythonEnvRepositoryLocation,
    RepositoryLocation,
)
from .represented import RepresentedPipeline
from .selector import PipelineSelector, RepositorySelector, ScheduleSelector



from dagster import check
from dagster.core.host_representation import (
    ExternalRepository,
    ExternalRepositoryData,
    RepositoryHandle,
    RepositoryLocationHandle,
)
from dagster.core.origin import RepositoryGrpcServerOrigin, RepositoryPythonOrigin

from .utils import execute_unary_api_cli_command


def sync_get_external_repositories(repository_location_handle):
    check.inst_param(
        repository_location_handle, 'repository_location_handle', RepositoryLocationHandle,
    )

    repos = []

    for _, pointer in repository_location_handle.repository_code_pointer_dict.items():
        external_repository_data = check.inst(
            execute_unary_api_cli_command(
                repository_location_handle.executable_path,
                'repository',
                RepositoryPythonOrigin(repository_location_handle.executable_path, pointer),
            ),
            ExternalRepositoryData,
        )
        repos.append(
            ExternalRepository(
                external_repository_data,
                RepositoryHandle(
                    repository_name=external_repository_data.name,
                    repository_location_handle=repository_location_handle,
                ),
            )
        )

    return repos


def sync_get_external_repositories_grpc(api_client, repository_location_handle):
    check.inst_param(
        repository_location_handle, 'repository_location_handle', RepositoryLocationHandle
    )

    repos = []
    for repository_name in repository_location_handle.repository_names:
        external_repository_data = check.inst(
            api_client.external_repository(
                repository_grpc_server_origin=RepositoryGrpcServerOrigin(
                    repository_location_handle.host,
                    repository_location_handle.port,
                    repository_location_handle.socket,
                    repository_name,
                )
            ),
            ExternalRepositoryData,
        )
        repos.append(
            ExternalRepository(
                external_repository_data,
                RepositoryHandle(
                    repository_name=external_repository_data.name,
                    repository_location_handle=repository_location_handle,
                ),
            )
        )
    return repos



from dagster import check
from dagster.config.field_utils import Permissive

from .config_schema import dagster_type_loader
from .dagster_type import DagsterType, PythonObjectDagsterType, resolve_dagster_type


@dagster_type_loader(Permissive())
def _dict_input(_context, value):
    return value


PythonDict = PythonObjectDagsterType(
    dict,
    'PythonDict',
    loader=_dict_input,
    description='''Represents a python dictionary to pass between solids''',
)


class _TypedPythonDict(DagsterType):
    def __init__(self, key_type, value_type):
        self.key_type = check.inst_param(key_type, 'key_type', DagsterType)
        self.value_type = check.inst_param(value_type, 'value_type', DagsterType)
        super(_TypedPythonDict, self).__init__(
            key='TypedPythonDict.{}.{}'.format(key_type.key, value_type.key),
            name=None,
            type_check_fn=self.type_check_method,
        )

    def type_check_method(self, context, value):
        from dagster.core.definitions.events import TypeCheck

        if not isinstance(value, dict):
            return TypeCheck(
                success=False,
                description='Value should be a dict, got a {value_type}'.format(
                    value_type=type(value)
                ),
            )

        for key, value in value.items():
            key_check = self.key_type.type_check(context, key)
            if not key_check.success:
                return key_check
            value_check = self.value_type.type_check(context, value)
            if not value_check.success:
                return value_check

        return TypeCheck(success=True)

    @property
    def display_name(self):
        return 'Dict[{key},{value}]'.format(
            key=self.key_type.display_name, value=self.value_type.display_name
        )

    @property
    def inner_types(self):
        return [self.key_type, self.value_type]

    @property
    def type_param_keys(self):
        return [self.key_type.key, self.value_type.key]


def create_typed_runtime_dict(key_dagster_type, value_dagster_type):
    key_type = resolve_dagster_type(key_dagster_type)
    value_type = resolve_dagster_type(value_dagster_type)

    return _TypedPythonDict(key_type, value_type)


class DagsterDictApi(object):
    def __getitem__(self, *args):
        check.param_invariant(len(args[0]) == 2, 'args', 'Must be two parameters')
        return create_typed_runtime_dict(args[0][0], args[0][1])


Dict = DagsterDictApi()



from abc import ABCMeta, abstractmethod, abstractproperty

import six

from dagster import check
from dagster.config.evaluate_value_result import EvaluateValueResult
from dagster.config.validate import process_config
from dagster.core.errors import DagsterConfigMappingFunctionError, user_code_error_boundary


class IConfigMappable(six.with_metaclass(ABCMeta)):
    @abstractproperty
    def _configured_config_mapping_fn(self):
        raise NotImplementedError()

    @abstractproperty
    def config_schema(self):
        raise NotImplementedError()

    @abstractmethod
    def configured(self, config_or_config_fn, config_schema=None, **kwargs):
        '''
        Wraps this object in an object of the same type that provides configuration to the inner
        object.

        Args:
            config_or_config_fn (Union[Any, Callable[[Any], Any]]): Either (1) Run configuration
                that fully satisfies this object's config schema or (2) A function that accepts run
                configuration and returns run configuration that fully satisfies this object's
                config schema.  In the latter case, config_schema must be specified.  When
                passing a function, it's easiest to use :py:func:`configured`.
            config_schema (ConfigSchema): If config_or_config_fn is a function, the config schema
                that its input must satisfy.
            **kwargs: Arbitrary keyword arguments that will be passed to the initializer of the
                returned object.

        Returns (IConfigMappable): A configured version of this object.
        '''
        raise NotImplementedError()

    def apply_config_mapping(self, config):
        '''
        Applies user-provided config mapping functions to the given configuration and validates the
        results against the respective config schema.

        Expects incoming config to be validated and have fully-resolved values (StringSource values
        resolved, Enum types hydrated, etc.) via process_config() during EnvironmentConfig
        construction and CompositeSolid config mapping.

        Args:
            config (Any): A validated and resolved configuration dictionary matching this object's
            config_schema

        Returns (EvaluateValueResult):
            If successful, the value is a validated and resolved configuration dictionary for the
            innermost wrapped object after applying the config mapping transformation function.
        '''
        # If there is no __configured_config_mapping_fn, this is the innermost resource (base case),
        # so we aren't responsible for validating against anything farther down. Returns an EVR for
        # type consistency with wrapped_config_mapping_fn.
        return (
            self._configured_config_mapping_fn(config)
            if self._configured_config_mapping_fn
            else EvaluateValueResult.for_value(config)
        )

    def _get_user_code_error_str_lambda(self):
        return lambda: (
            'The config mapping function on a `configured` {} has thrown an unexpected '
            'error during its execution.'
        ).format(self.__class__.__name__)

    def _get_wrapped_config_mapping_fn(self, config_or_config_fn, config_schema):
        '''
        Returns a config mapping helper function that will be stored on the child `IConfigMappable`
        under `_configured_config_mapping_fn`. Encapsulates the recursiveness of the `configurable`
        pattern by returning a closure that invokes `self.apply_config_mapping` (belonging to this
        parent object) on the mapping function or static config passed into this method.
        '''
        # Provide either a config mapping function with noneable schema...
        if callable(config_or_config_fn):
            config_fn = config_or_config_fn
        else:  # or a static config object (and no schema).
            check.invariant(
                config_schema is None,
                'When non-callable config is given, config_schema must be None',
            )

            def config_fn(_):  # Stub a config mapping function from the static config object.
                return config_or_config_fn

        def wrapped_config_mapping_fn(validated_and_resolved_config):
            '''
            Given validated and resolved configuration for this IConfigMappable, applies the
            provided config mapping function, validates its output against the inner resource's
            config_schema, and recursively invoked the `apply_config_mapping` method on the resource
            '''
            check.dict_param(validated_and_resolved_config, 'validated_and_resolved_config')
            with user_code_error_boundary(
                DagsterConfigMappingFunctionError, self._get_user_code_error_str_lambda()
            ):
                mapped_config = {
                    'config': config_fn(validated_and_resolved_config.get('config', {}))
                }
            # Validate mapped_config against the inner resource's config_schema (on self).
            config_evr = process_config({'config': self.config_schema or {}}, mapped_config)
            if config_evr.success:
                return self.apply_config_mapping(config_evr.value)  # Recursive step
            else:
                return config_evr  # Bubble up the errors

        return wrapped_config_mapping_fn


def configured(configurable, config_schema=None, **kwargs):
    '''
    A decorator that makes it easy to create a function-configured version of an object.

    Args:
        configurable (IConfigMappable): An object that can be configured.
        config_schema (ConfigSchema): The config schema that the inputs to the decorated function
            must satisfy.
        **kwargs: Arbitrary keyword arguments that will be passed to the initializer of the returned
            object.

    Returns (Callable[[Union[Any, Callable[[Any], Any]]], IConfigMappable])

    Examples:

        .. code-block:: python

            dev_s3 = configured(s3_resource)({'bucket': 'dev'})

            @configured(s3_resource):
            def dev_s3(_):
                return {'bucket': 'dev'}

            @configured(s3_resource, {'bucket_prefix', str}):
            def dev_s3(config):
                return {'bucket': config['bucket_prefix'] + 'dev'}
    '''

    check.inst_param(configurable, 'configurable', IConfigMappable)

    def _configured(config_or_config_fn):
        return configurable.configured(
            config_schema=config_schema, config_or_config_fn=config_or_config_fn, **kwargs
        )

    return _configured



from collections import namedtuple
from enum import Enum

from dagster import check
from dagster.core.code_pointer import (
    CodePointer,
    FileCodePointer,
    ModuleCodePointer,
    PackageCodePointer,
)
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.instance.ref import InstanceRef
from dagster.core.origin import (
    PipelineOrigin,
    PipelinePythonOrigin,
    RepositoryOrigin,
    RepositoryPythonOrigin,
)
from dagster.serdes import whitelist_for_serdes
from dagster.utils.error import SerializableErrorInfo


class LoadableTargetOrigin(
    namedtuple(
        'LoadableTargetOrigin',
        'executable_path python_file module_name working_directory attribute',
    )
):
    def __new__(
        cls,
        executable_path=None,
        python_file=None,
        module_name=None,
        working_directory=None,
        attribute=None,
    ):
        return super(LoadableTargetOrigin, cls).__new__(
            cls,
            executable_path=check.opt_str_param(executable_path, 'executable_path'),
            python_file=check.opt_str_param(python_file, 'python_file'),
            module_name=check.opt_str_param(module_name, 'module_name'),
            working_directory=check.opt_str_param(working_directory, 'working_directory'),
            attribute=check.opt_str_param(attribute, 'attribute'),
        )

    @staticmethod
    def from_python_origin(repository_python_origin):
        check.inst_param(
            repository_python_origin, 'repository_python_origin', RepositoryPythonOrigin
        )
        executable_path = repository_python_origin.executable_path
        code_pointer = repository_python_origin.code_pointer
        if isinstance(code_pointer, FileCodePointer):
            return LoadableTargetOrigin(
                executable_path=executable_path,
                python_file=code_pointer.python_file,
                attribute=code_pointer.fn_name,
                working_directory=code_pointer.working_directory,
            )
        elif isinstance(code_pointer, ModuleCodePointer):
            return LoadableTargetOrigin(
                executable_path=executable_path,
                module_name=code_pointer.module,
                attribute=code_pointer.fn_name,
            )
        elif isinstance(code_pointer, PackageCodePointer):
            return LoadableTargetOrigin(
                executable_path=executable_path,
                module_name=code_pointer.module,
                attribute=code_pointer.attribute,
            )
        else:
            raise DagsterInvariantViolationError(
                "Unexpected code pointer {code_pointer_name}".format(
                    code_pointer_name=type(code_pointer).__name__
                )
            )


@whitelist_for_serdes
class ExecutionPlanSnapshotArgs(
    namedtuple(
        '_ExecutionPlanSnapshotArgs',
        'pipeline_origin solid_selection run_config mode step_keys_to_execute pipeline_snapshot_id',
    )
):
    def __new__(
        cls,
        pipeline_origin,
        solid_selection,
        run_config,
        mode,
        step_keys_to_execute,
        pipeline_snapshot_id,
    ):
        return super(ExecutionPlanSnapshotArgs, cls).__new__(
            cls,
            pipeline_origin=check.inst_param(pipeline_origin, 'pipeline_origin', PipelineOrigin),
            solid_selection=check.opt_list_param(solid_selection, 'solid_selection', of_type=str),
            run_config=check.dict_param(run_config, 'run_config'),
            mode=check.str_param(mode, 'mode'),
            step_keys_to_execute=check.opt_list_param(
                step_keys_to_execute, 'step_keys_to_execute', of_type=str
            ),
            pipeline_snapshot_id=check.str_param(pipeline_snapshot_id, 'pipeline_snapshot_id'),
        )


@whitelist_for_serdes
class ExecuteRunArgs(namedtuple('_ExecuteRunArgs', 'pipeline_origin pipeline_run_id instance_ref')):
    def __new__(cls, pipeline_origin, pipeline_run_id, instance_ref):
        return super(ExecuteRunArgs, cls).__new__(
            cls,
            pipeline_origin=check.inst_param(pipeline_origin, 'pipeline_origin', PipelineOrigin),
            pipeline_run_id=check.str_param(pipeline_run_id, 'pipeline_run_id'),
            instance_ref=check.opt_inst_param(instance_ref, 'instance_ref', InstanceRef),
        )


@whitelist_for_serdes
class ExecuteStepArgs(
    namedtuple(
        '_ExecuteStepArgs',
        'pipeline_origin pipeline_run_id instance_ref mode step_keys_to_execute run_config retries_dict',
    )
):
    def __new__(
        cls,
        pipeline_origin,
        pipeline_run_id,
        instance_ref=None,
        mode=None,
        step_keys_to_execute=None,
        run_config=None,
        retries_dict=None,
    ):
        return super(ExecuteStepArgs, cls).__new__(
            cls,
            pipeline_origin=check.inst_param(
                pipeline_origin, 'pipeline_origin', PipelinePythonOrigin
            ),
            pipeline_run_id=check.str_param(pipeline_run_id, 'pipeline_run_id'),
            instance_ref=check.opt_inst_param(instance_ref, 'instance_ref', InstanceRef),
            mode=check.opt_str_param(mode, 'mode'),
            step_keys_to_execute=check.opt_list_param(
                step_keys_to_execute, 'step_keys_to_execute', of_type=str
            ),
            run_config=check.opt_dict_param(run_config, 'run_config'),
            retries_dict=check.opt_dict_param(retries_dict, 'retries_dict'),
        )


@whitelist_for_serdes
class LoadableRepositorySymbol(
    namedtuple('_LoadableRepositorySymbol', 'repository_name attribute')
):
    def __new__(cls, repository_name, attribute):
        return super(LoadableRepositorySymbol, cls).__new__(
            cls,
            repository_name=check.str_param(repository_name, 'repository_name'),
            attribute=check.str_param(attribute, 'attribute'),
        )


@whitelist_for_serdes
class ListRepositoriesResponse(
    namedtuple(
        '_ListRepositoriesResponse',
        'repository_symbols executable_path repository_code_pointer_dict',
    )
):
    def __new__(
        cls, repository_symbols, executable_path=None, repository_code_pointer_dict=None,
    ):
        return super(ListRepositoriesResponse, cls).__new__(
            cls,
            repository_symbols=check.list_param(
                repository_symbols, 'repository_symbols', of_type=LoadableRepositorySymbol
            ),
            # These are currently only used by the GRPC Repository Location, but
            # we will need to migrate the rest of the repository locations to use this.
            executable_path=check.opt_str_param(executable_path, 'executable_path'),
            repository_code_pointer_dict=check.opt_dict_param(
                repository_code_pointer_dict,
                'repository_code_pointer_dict',
                key_type=str,
                value_type=CodePointer,
            ),
        )


@whitelist_for_serdes
class ListRepositoriesInput(
    namedtuple('_ListRepositoriesInput', 'module_name python_file working_directory attribute')
):
    def __new__(cls, module_name, python_file, working_directory, attribute):
        check.invariant(not (module_name and python_file), 'Must set only one')
        check.invariant(module_name or python_file, 'Must set at least one')
        return super(ListRepositoriesInput, cls).__new__(
            cls,
            module_name=check.opt_str_param(module_name, 'module_name'),
            python_file=check.opt_str_param(python_file, 'python_file'),
            working_directory=check.opt_str_param(working_directory, 'working_directory'),
            attribute=check.opt_str_param(attribute, 'attribute'),
        )


@whitelist_for_serdes
class PartitionArgs(
    namedtuple('_PartitionArgs', 'repository_origin partition_set_name partition_name')
):
    def __new__(cls, repository_origin, partition_set_name, partition_name):
        return super(PartitionArgs, cls).__new__(
            cls,
            repository_origin=check.inst_param(
                repository_origin, 'repository_origin', RepositoryOrigin
            ),
            partition_set_name=check.str_param(partition_set_name, 'partition_set_name'),
            partition_name=check.str_param(partition_name, 'partition_name'),
        )


@whitelist_for_serdes
class PartitionNamesArgs(namedtuple('_PartitionNamesArgs', 'repository_origin partition_set_name')):
    def __new__(cls, repository_origin, partition_set_name):
        return super(PartitionNamesArgs, cls).__new__(
            cls,
            repository_origin=check.inst_param(
                repository_origin, 'repository_origin', RepositoryOrigin
            ),
            partition_set_name=check.str_param(partition_set_name, 'partition_set_name'),
        )


@whitelist_for_serdes
class PipelineSubsetSnapshotArgs(
    namedtuple('_PipelineSubsetSnapshotArgs', 'pipeline_origin solid_selection')
):
    def __new__(cls, pipeline_origin, solid_selection):
        return super(PipelineSubsetSnapshotArgs, cls).__new__(
            cls,
            pipeline_origin=check.inst_param(pipeline_origin, 'pipeline_origin', PipelineOrigin),
            solid_selection=check.list_param(solid_selection, 'solid_selection', of_type=str)
            if solid_selection
            else None,
        )


@whitelist_for_serdes
class ScheduleExecutionDataMode(Enum):
    # Just return the schedule data and tags, don't check whether we should excute the schedule or return it
    PREVIEW = 'PREVIEW'

    # Return schedule data, tags, and whether we should execute the schedule
    LAUNCH_SCHEDULED_EXECUTION = 'LAUNCH_SCHEDULED_EXECUTION'


@whitelist_for_serdes
class ExternalScheduleExecutionArgs(
    namedtuple(
        '_ExternalScheduleExecutionArgs',
        'repository_origin instance_ref schedule_name schedule_execution_data_mode',
    )
):
    def __new__(cls, repository_origin, instance_ref, schedule_name, schedule_execution_data_mode):
        return super(ExternalScheduleExecutionArgs, cls).__new__(
            cls,
            repository_origin=check.inst_param(
                repository_origin, 'repository_origin', RepositoryOrigin
            ),
            instance_ref=check.inst_param(instance_ref, 'instance_ref', InstanceRef),
            schedule_name=check.str_param(schedule_name, 'schedule_name'),
            schedule_execution_data_mode=check.inst_param(
                schedule_execution_data_mode,
                'schedule_execution_data_mode',
                ScheduleExecutionDataMode,
            ),
        )


@whitelist_for_serdes
class ShutdownServerResult(namedtuple('_ShutdownServerResult', 'success serializable_error_info')):
    def __new__(cls, success, serializable_error_info):
        return super(ShutdownServerResult, cls).__new__(
            cls,
            success=check.bool_param(success, 'success'),
            serializable_error_info=check.opt_inst_param(
                serializable_error_info, 'serializable_error_info', SerializableErrorInfo
            ),
        )


@whitelist_for_serdes
class CancelExecutionRequest(namedtuple('_CancelExecutionRequest', 'run_id')):
    def __new__(cls, run_id):
        return super(CancelExecutionRequest, cls).__new__(
            cls, run_id=check.str_param(run_id, 'run_id'),
        )


@whitelist_for_serdes
class CancelExecutionResult(
    namedtuple('_CancelExecutionResult', 'success message serializable_error_info')
):
    def __new__(cls, success, message, serializable_error_info):
        return super(CancelExecutionResult, cls).__new__(
            cls,
            success=check.bool_param(success, 'success'),
            message=check.opt_str_param(message, 'message'),
            serializable_error_info=check.opt_inst_param(
                serializable_error_info, 'serializable_error_info', SerializableErrorInfo
            ),
        )


@whitelist_for_serdes
class CanCancelExecutionRequest(namedtuple('_CanCancelExecutionRequest', 'run_id')):
    def __new__(cls, run_id):
        return super(CanCancelExecutionRequest, cls).__new__(
            cls, run_id=check.str_param(run_id, 'run_id'),
        )


@whitelist_for_serdes
class CanCancelExecutionResult(namedtuple('_CancelExecutionResult', 'can_cancel')):
    def __new__(cls, can_cancel):
        return super(CanCancelExecutionResult, cls).__new__(
            cls, can_cancel=check.bool_param(can_cancel, 'can_cancel'),
        )


@whitelist_for_serdes
class StartRunResult(namedtuple('_StartRunResult', 'success message serializable_error_info')):
    def __new__(cls, success, message, serializable_error_info):
        return super(StartRunResult, cls).__new__(
            cls,
            success=check.bool_param(success, 'success'),
            message=check.opt_str_param(message, 'message'),
            serializable_error_info=check.opt_inst_param(
                serializable_error_info, 'serializable_error_info', SerializableErrorInfo
            ),
        )


@whitelist_for_serdes
class GetCurrentImageResult(
    namedtuple('_GetCurrentImageResult', 'current_image serializable_error_info')
):
    def __new__(cls, current_image, serializable_error_info):
        return super(GetCurrentImageResult, cls).__new__(
            cls,
            current_image=check.opt_str_param(current_image, 'current_image'),
            serializable_error_info=check.opt_inst_param(
                serializable_error_info, 'serializable_error_info', SerializableErrorInfo
            ),
        )
